/*
 * Filesystem client functions.
 *
 * Copyright (C) 2018, 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/sys/err.h>

#include <fcntl.h>
#include <sys/types.h>
#include <unistd.h>

#include "client.h"
#include "client_fcntl.h" /* NOTE: Move to systypes! */
#include "fsdesc.h"
#include "fsdesc_client.h"
#include "fsops.h"
#include <ipc/mem_ipc.h>



/* Default size of mapped memory region. */

const int DEFAULT_BUFFER_SIZE = 4096;



/* Obtain a filesystem reference. */

file_descriptor_t *client_getfs(file_descriptor_t *desc, const char *name)
{
  if (desc == NULL)
    return NULL;

  if (fsdesc_init_cap(desc, "fs") == NULL)
    return NULL;

  /* Obtain a context. */

  if (fs_context(desc))
  {
    client_close(desc);
    return NULL;
  }

  /* Write the filename to the buffer provided by the context. */

  fsdesc_set_name(&desc->cobj, name);

  /* Deallocate the object if the operation failed. */

  if (fs_context_getfs(desc))
  {
    client_close(desc);
    return NULL;
  }

  /* Release the context. */

  fsdesc_deallocate_context(desc);

  return desc;
}



/* Mount a filesystem. */

long client_mount(file_descriptor_t *desc, const char *devname, const char *mountpath)
{
  long result;

  /* Open the device as a file, obtaining its capability. */

  desc = client_getfs(desc, devname);

  if (desc == NULL)
    return -L4_EIO;

  /* Obtain a context. */

  if (fs_context(desc))
  {
    client_close(desc);
    return -L4_EIO;
  }

  /* Write the mountpoint to the buffer provided by the context. */

  fsdesc_set_name(&desc->cobj, mountpath);

  /* Employ the descriptor to present the capability to the filesystem. */

  result = fs_context_mount(desc);

  client_close(desc);

  return result;
}



/* Open a filesystem object. */

file_descriptor_t *client_open(file_descriptor_t *desc, const char *name,
                               int flags, size_t buffer_size)
{
  if (desc == NULL)
    return NULL;

  if (fsdesc_init_cap(desc, "fs") == NULL)
    return NULL;

  /* Obtain a context. */

  if (fs_context(desc))
  {
    client_close(desc);
    return NULL;
  }

  /* Write the filename to the buffer provided by the context. */

  fsdesc_set_name(&desc->cobj, name);

  /* Deallocate the object if the operation failed. */

  if (fs_context_open(desc, flags))
  {
    client_close(desc);
    return NULL;
  }

  if (client_open_desc(desc, flags, buffer_size))
    return desc;
  else
    return NULL;
}



/* Open a device as a file. */

file_descriptor_t *client_open_device(file_descriptor_t *desc, const char *cap,
                                      int flags, size_t buffer_size)
{
  if (desc == NULL)
    return NULL;

  if (fsdesc_init_cap(desc, cap) == NULL)
    return NULL;

  /* Obtain a context. */

  if (fs_context(desc))
  {
    client_close(desc);
    return NULL;
  }

  /* Deallocate the object if the operation failed. */

  if (fs_context_open(desc, flags))
  {
    client_close(desc);
    return NULL;
  }

  if (client_open_desc(desc, flags, buffer_size))
    return desc;
  else
    return NULL;
}



/* Complete the initialisation of a descriptor. */

int client_open_desc(file_descriptor_t *desc, int flags, size_t buffer_size)
{
  /* Release the context. */

  fsdesc_deallocate_context(desc);

  /* Record the opening flags. */

  desc->flags = flags;

  /* Initially, memory-map the start of the file. If the memory mapping fails,
     the dataspace address for the object will be zero. */

  desc->obj.buffer.addr = client_mmap(desc, 0, buffer_size ? buffer_size
                                                           : DEFAULT_BUFFER_SIZE);

  /* Files can always be mapped using mmap. */

  desc->can_mmap = 1;

  /* Return with success if a buffer was assigned. */

  if (desc->obj.buffer.addr)
    return 1;

  /* Release all resources. */

  client_close(desc);
  return 0;
}



/* Open a pipe object. */

long client_pipe(file_descriptor_t *reader, file_descriptor_t *writer)
{
  long err;

  if ((reader == NULL) || (writer == NULL))
    return -L4_EINVAL;

  if (fsdesc_init_cap(reader, "pipes") == NULL)
    return -L4_ENOMEM;

  if (fsdesc_init_cap(writer, "pipes") == NULL)
    return -L4_ENOMEM;

  err = fs_pipe(reader->server, reader, writer);
  if (err)
    return err;

  if ((client_current_region(writer) == NULL) ||
      (client_current_region(reader) == NULL))
  {
    client_close(reader);
    client_close(writer);
    return -L4_ENOMEM;
  }

  return L4_EOK;
}



/* Start synchronisation with the object. */

long client_begin(file_descriptor_t *desc)
{
  return fs_begin(desc);
}

long client_begin_first(file_descriptor_t *desc)
{
  return fs_begin_first(desc);
}



/* End synchronisation with the object. */

long client_end(file_descriptor_t *desc)
{
  return fs_end(desc);
}



/* Close a filesystem object. */

void client_close(file_descriptor_t *desc)
{
  fsdesc_deallocate(desc);
}



/* Flush data conditionally to the filesystem object. */

static long _flush(file_descriptor_t *desc, size_t position)
{
  long err;

  /* Where the position is outside the current region, re-map. */

  if ((position < desc->obj.start_pos) ||
      (position >= desc->obj.start_pos + desc->obj.buffer.size))
  {
    if (desc->can_mmap)
    {
      if (!client_mmap(desc, position, position + desc->obj.buffer.size))
        return -L4_EIO;
    }

    /* Strict conditions for region navigation in pipes. */

    else if ((position != desc->obj.start_pos + desc->obj.buffer.size) ||
             (client_next_region(desc) == NULL))
           return -L4_EIO;
  }

  /* Otherwise, flush any written data in the current region. */

  else
  {
    err = fs_flush(desc);
    if (err)
      return err;

    desc->written = 0;
  }

  return L4_EOK;
}



/* Flush data explicitly to the filesystem object. */

long client_flush(file_descriptor_t *desc)
{
  /* Flush and retain most buffer settings. */

  long err = fs_flush(desc);

  if (!err)
    desc->written = 0;

  return err;
}



/* Read the metadata for a file. */

long client_fstat(file_descriptor_t *desc, struct stat *statbuf)
{
  long err = fs_fstat(desc, statbuf);

  if (err)
    return err;

  return L4_EOK;
}



/* Attach a file-like resource as a dataspace. */

static void *_attach_region(file_descriptor_t *desc)
{
  if (desc->obj.buffer.addr)
    ipc_detach_dataspace(desc->obj.buffer.addr);

  ipc_attach_dataspace(desc->ref, desc->obj.buffer.size,
                       (void **) &desc->obj.buffer.addr);

  return desc->obj.buffer.addr;
}



/* Map a memory region to a file. */

void *client_mmap(file_descriptor_t *desc, off_t position, size_t length)
{
  if (fs_mmap(desc, position, length))
    return NULL;

  return _attach_region(desc);
}



/* Obtain the current region of a pipe. */

void *client_current_region(file_descriptor_t *desc)
{
  if (fs_current_region(desc))
    return NULL;

  return _attach_region(desc);
}



/* Obtain the next region of a pipe. */

void *client_next_region(file_descriptor_t *desc)
{
  if (fs_next_region(desc))
    return NULL;

  return _attach_region(desc);
}



/* Read from the filesystem object into the buffer provided. */

ssize_t client_read(file_descriptor_t *desc, void *buf, size_t count)
{
  size_t to_transfer, total = 0;

  /* Amount available in the descriptor buffer already. */

  size_t available = fs_get_available(desc);

  while (count > 0)
  {
    /* If there is no data, try and obtain more data. */

    if (!available)
    {
      /* Flush any unwritten data, preparing to read from the file position at
         the end of the data. */

      if (_flush(desc, fs_get_data_end_position(desc)))
        break;

      available = fs_get_available(desc);

      if (!available)
        break;
    }

    /* Transfer data into the supplied buffer. */

    to_transfer = available <= count ? available : count;

    fs_read_desc(desc, (char *) buf, to_transfer);

    /* Update counters. */

    available -= to_transfer;

    count -= to_transfer;
    total += to_transfer;

    buf = ((char *) buf + to_transfer);
  }

  return total;
}



#define DIRENT_CORE_SIZE (sizeof(struct dirent) - sizeof(((struct dirent *) 0)->d_name))
#define DIRENT_STAT_CORE_SIZE (DIRENT_CORE_SIZE + sizeof(struct stat))
#define DIRENT_STAT_SIZE(result) (result->d_reclen + sizeof(struct stat))



/* Test for an available directory entry in the buffer. */

static struct dirent *_get_dirent(file_descriptor_t *desc)
{
  /* Amount available in the descriptor buffer already. */

  size_t available = fs_get_available(desc);

  /* First available record, if any. */

  struct dirent *result = (struct dirent *) fs_get_data_start(desc);

  /* Ensure enough data for one record.
     NOTE: Records longer than a page are not supported. */

  if (available < DIRENT_CORE_SIZE)
    return NULL;

  if ((desc->flags & OPENDIR_STAT) && (available < DIRENT_STAT_CORE_SIZE))
    return NULL;

  if (available < result->d_reclen)
    return NULL;

  if ((desc->flags & OPENDIR_STAT) && (available < DIRENT_STAT_SIZE(result)))
    return NULL;

  return result;
}



/* Read a directory entry directly from the transfer buffer. */

struct dirent *client_readdir(file_descriptor_t *desc)
{
  struct dirent *result = _get_dirent(desc);
  size_t position;
  int size;

  if (result == NULL)
  {
    position = fs_get_data_start_position(desc);

    /* Remap the dataspace. */

    if (!client_mmap(desc, position, position + desc->obj.buffer.size))
      return NULL;

    /* Check available data again. */

    result = _get_dirent(desc);

    if (result == NULL)
      return NULL;
  }

  /* Test for a null record. */

  if (!result->d_reclen)
    return NULL;

  /* Reference the current record and advance the file position to the next
     (which may be the end of the buffer data, requiring more to be fetched). */

  size = result->d_reclen;

  if (desc->flags & OPENDIR_STAT)
    size += sizeof(struct stat);

  client_seek(desc, size, SEEK_CUR);

  return result;
}



/* Ensure that the buffer can provide the needed data. */

long client_ensure_read_position(file_descriptor_t *desc, off_t position)
{
  /* Load region data at the desired position if appropriate. */

  if (fs_no_available_data(desc) || !fs_have_position(desc, position))
  {
    if (client_seek(desc, position, SEEK_SET) != position)
      return -L4_EIO;
  }

  /* Otherwise, position the data offset if the position resides within the
     region. */

  else
    desc->data_start = position - desc->obj.start_pos;

  return L4_EOK;
}



off_t client_seek(file_descriptor_t *desc, off_t offset, int whence)
{
  size_t position, current = fs_get_data_start_position(desc), change;

  switch (whence)
  {
    case SEEK_SET:
      position = offset;
      break;

    case SEEK_CUR:
      position = current + offset;
      break;

    case SEEK_END:
      position = desc->obj.size + offset;
      break;

    default:
      /* NOTE: Set errno to EINVAL. */
      return -1;
  }

  /* Retain the current position if unchanged. */

  if (position == current)
    return position;

  /* Move forward in the file. */

  if (position > current)
  {
    change = position - current;

    /* Move towards the end of available data.
       Request new data if not enough is available. */

    if (change <= fs_get_available(desc))
    {
      desc->data_start += change;
      return position;
    }
  }

  /* Move backward in the file. */

  else
  {
    change = current - position;

    /* Move towards the start of available data.
       Request new data if moving beyond the start of the data. */

    if (change <= desc->data_start)
    {
      desc->data_start -= change;
      return position;
    }
  }

  /* Handle unwritten data and reset the buffer for reading. */

  _flush(desc, position);

  return position;
}



long client_tell(file_descriptor_t *desc)
{
  return fs_get_data_start_position(desc);
}



/* Write to the filesystem object from the buffer provided. */

ssize_t client_write(file_descriptor_t *desc, const void *buf, size_t count)
{
  size_t to_transfer, total = 0;

  /* Space remaining in the descriptor buffer. */

  size_t space = fs_get_space(desc);

  while (count > 0)
  {
    /* If no space is available, try and send data, reset the buffer. */

    if (!space)
    {
      /* Flush any unwritten data and continue writing from the start of the
         data. */

      if (_flush(desc, fs_get_data_start_position(desc)))
        break;

      space = fs_get_space(desc);
    }

    /* Transfer data into the supplied buffer. */

    to_transfer = space <= count ? space : count;

    fs_write_desc(desc, (char *) buf, to_transfer);

    /* Update counters. */

    space -= to_transfer;

    count -= to_transfer;
    total += to_transfer;

    buf = ((char *) buf + to_transfer);
  }

  return total;
}



/* Yield to another client of the object. */

long client_yield(file_descriptor_t *desc)
{
  return fs_yield(desc);
}
